package be.iminds.ilabt.jfed.restapi.json;

import be.iminds.ilabt.jfed.restapi.json.util.JsonGeniUrnDeserializer;
import be.iminds.ilabt.jfed.restapi.json.util.JsonGeniUrnSerializer;
import be.iminds.ilabt.jfed.restapi.json.util.JsonTimeStampRFC3339Deserializer;
import be.iminds.ilabt.jfed.restapi.json.util.JsonTimeStampRFC3339Serializer;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.util.jsonld.JsonLdObjectsMetaData;
import be.iminds.ilabt.util.jsonld.impl.PrimaryIdObject;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.net.URL;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

@JsonInclude(JsonInclude.Include.NON_NULL)
@JsonIgnoreProperties("@context")
@JsonLdObjectsMetaData.JsonLdObjectInfo(pathName = "experiment", builderClass = ExperimentBuilder.class,
uriType = JsonLdObjectsMetaData.UriType.ID_BASED, idType = JsonLdObjectsMetaData.IdType.PRIMARY)
public class Experiment extends PrimaryIdObject<Long> {
    private static final Logger LOG = LoggerFactory.getLogger(Experiment.class);

    public enum Status { STARTING, READY, FAILED, NONEXISTING };

    private final Timestamp expire;
    private final Timestamp created;
    private final String requestRspec;
    private final String manifestRspec;
    private final String speaksForCredential;
    private final GeniUrn speakingFor;
    private final String project;
    private final String sliceName;
//    private final Server am;
    private final URL am;
    private final List<GeniUrn> sharedWith;
    private final GeniUrn owner;
    private final GeniUrn sliceUrn;

    private final Status status;
    private final StatusDetails statusDetails;

    @JsonCreator
    Experiment(@JsonProperty("id") Long id,
//               @JsonProperty("name") String name,
               @JsonProperty("expire") Timestamp expire,
               @JsonProperty("created") Timestamp created,
               @JsonProperty("requestRspec") String requestRspec,
               @JsonProperty("manifestRspec") String manifestRspec,
               @JsonProperty("speaksForCredential") String speaksForCredential,
               @JsonProperty("speakingFor") GeniUrn speakingFor,
               @JsonProperty("project") String project,
               @JsonProperty("sliceName") String sliceName,
               @JsonProperty("am") URL am,
               @JsonProperty("sharedWith") List<GeniUrn> sharedWith,
               @JsonProperty("owner") GeniUrn owner,
               @JsonProperty("sliceUrn") GeniUrn sliceUrn,
               @JsonProperty("status") Status status,
               @JsonProperty("statusDetails") StatusDetails statusDetails,
               @JsonProperty("@id") URI uri) {
        this(id, expire, created, requestRspec, manifestRspec, speaksForCredential, speakingFor, project,
                sliceName, am, sharedWith, owner, sliceUrn, status, statusDetails, uri,
                JsonLdObjectsMetaData.mustSerializeAsEmbeddedObjectForJacksonConstructor());
    }

    Experiment(Long id,
               Timestamp expire,
               Timestamp created,
               String requestRspec,
               String manifestRspec,
               String speaksForCredential,
               GeniUrn speakingFor,
               String project,
               String sliceName,
               URL am,
               List<GeniUrn> sharedWith,
               GeniUrn owner,
               GeniUrn sliceUrn,
               Status status,
               StatusDetails statusDetails,
               URI uri,
               boolean serializeInfo_SerializeAsEmbeddedObject) {
        super(id, uri, serializeInfo_SerializeAsEmbeddedObject);
//        this.name = name;
        this.expire = expire;
        this.created = created;
        this.requestRspec = requestRspec;
        this.manifestRspec = manifestRspec;
        this.speaksForCredential = speaksForCredential;
        this.speakingFor = speakingFor;
        this.project = project;
        this.sliceName = sliceName;
        this.am = am;
        this.sharedWith = sharedWith == null ? null : Collections.unmodifiableList(new ArrayList<>(sharedWith));
        this.owner = owner;
        this.sliceUrn = sliceUrn;
        this.status = status;
        this.statusDetails = statusDetails;
    }


    @JsonProperty
    @JsonSerialize(using = JsonTimeStampRFC3339Serializer.class)
    @JsonDeserialize(using = JsonTimeStampRFC3339Deserializer.class)
    public Timestamp getExpire() {
        return expire;
    }

    @JsonProperty
    @JsonSerialize(using = JsonTimeStampRFC3339Serializer.class)
    @JsonDeserialize(using = JsonTimeStampRFC3339Deserializer.class)
    public Timestamp getCreated() {
        return created;
    }

    @JsonProperty
    public String getRequestRspec() {
        return requestRspec;
    }

    @JsonProperty
    public String getManifestRspec() {
        return manifestRspec;
    }

    @JsonProperty
    public String getSpeaksForCredential() {
        return speaksForCredential;
    }

    @JsonProperty
    @JsonSerialize(using = JsonGeniUrnSerializer.class)
    @JsonDeserialize(using = JsonGeniUrnDeserializer.class)
    public GeniUrn getSpeakingFor() {
        return speakingFor;
    }

    @JsonIgnore
    public String getSpeakingForAsString() {
        return speakingFor == null ? null : speakingFor.toString();
    }

    @JsonProperty
    public String getProject() {
        return project;
    }

    @JsonProperty
    public String getSliceName() {
        return sliceName;
    }

    @JsonProperty
    public URL getAm() {
        return am;
    }

    @JsonIgnore
    public String getAmAsString() {
        return am == null ? null : am.toExternalForm();
    }

    @JsonProperty
    @JsonSerialize(contentUsing = JsonGeniUrnSerializer.class)
    @JsonDeserialize(contentUsing = JsonGeniUrnDeserializer.class)
    public List<GeniUrn> getSharedWith() {
        return sharedWith;
    }

    @JsonIgnore
    public String getSharedWithAsSsv() {
        return sharedWith == null ? null : sharedWith.stream().map(GeniUrn::toString).collect(Collectors.joining(" "));
    }

    @JsonProperty
    @JsonSerialize(using = JsonGeniUrnSerializer.class)
    @JsonDeserialize(using = JsonGeniUrnDeserializer.class)
    public GeniUrn getOwner() {
        return owner;
    }

    @JsonIgnore
    public String getOwnerAsString() {
        return owner == null ? null : owner.toString();
    }

    @JsonProperty
    @JsonSerialize(using = JsonGeniUrnSerializer.class)
    @JsonDeserialize(using = JsonGeniUrnDeserializer.class)
    public GeniUrn getSliceUrn() {
        return sliceUrn;
    }

    @JsonIgnore
    public String getSliceUrnAsString() {
        return sliceUrn == null ? null : sliceUrn.toString();
    }

    @JsonProperty
    public Status getStatus() {
        return status;
    }
    @JsonIgnore
    public String getStatusAsString() {
        return status == null ? null : status.toString();
    }

    @JsonProperty
    public StatusDetails getStatusDetails() {
        return statusDetails;
    }

    @JsonIgnore
    public String getStatusDetailsAsJsonString() {
        if (statusDetails == null) return null;
        ObjectMapper mapper = new ObjectMapper();
        try {
            return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(statusDetails);
        } catch (JsonProcessingException e) {
            LOG.error("Exception converting "+StatusDetails.class.getSimpleName()+" to JSON", e);
            return "Exception converting "+StatusDetails.class.getSimpleName()+" to JSON: "+e.getMessage();
        }
    }

    @Override
    public String toString() {
        ObjectMapper mapper = new ObjectMapper();
        try {
            return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(this);
        } catch (JsonProcessingException e) {
            LOG.error("Exception converting "+getClass().getSimpleName()+" to JSON", e);
            return "Exception converting "+getClass().getSimpleName()+" to JSON: "+e.getMessage();
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties("@context")
    public static class StatusDetails {
        private final String globalSliverStatus;
        private final String allocationStatus;
        private final String operationalStatus;
        private final String jFedExperimentState;

        @JsonCreator
        public StatusDetails(@JsonProperty("globalSliverStatus") String globalSliverStatus,
                             @JsonProperty("allocationStatus") String allocationStatus,
                             @JsonProperty("operationalStatus") String operationalStatus,
                             @JsonProperty("jFedExperimentState") String jFedExperimentState) {
            this.globalSliverStatus = globalSliverStatus;
            this.allocationStatus = allocationStatus;
            this.operationalStatus = operationalStatus;
            this.jFedExperimentState = jFedExperimentState;
        }

        @JsonProperty
        public String getGlobalSliverStatus() {
            return globalSliverStatus;
        }

        @JsonProperty
        public String getAllocationStatus() {
            return allocationStatus;
        }

        @JsonProperty
        public String getOperationalStatus() {
            return operationalStatus;
        }

        @JsonProperty
        public String getjFedExperimentState() {
            return jFedExperimentState;
        }

        @Override
        public String toString() {
            ObjectMapper mapper = new ObjectMapper();
            try {
                return mapper.writerWithDefaultPrettyPrinter().writeValueAsString(this);
            } catch (JsonProcessingException e) {
                LOG.error("Exception converting "+getClass().getSimpleName()+" to JSON", e);
                return "Exception converting "+getClass().getSimpleName()+" to JSON: "+e.getMessage();
            }
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            StatusDetails that = (StatusDetails) o;

            if (globalSliverStatus != null ? !globalSliverStatus.equals(that.globalSliverStatus) : that.globalSliverStatus != null)
                return false;
            if (allocationStatus != null ? !allocationStatus.equals(that.allocationStatus) : that.allocationStatus != null)
                return false;
            if (jFedExperimentState != null ? !jFedExperimentState.equals(that.jFedExperimentState) : that.jFedExperimentState != null) return false;
            if (operationalStatus != null ? !operationalStatus.equals(that.operationalStatus) : that.operationalStatus != null)
                return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = allocationStatus != null ? allocationStatus.hashCode() : 0;
            result = 31 * result + (globalSliverStatus != null ? globalSliverStatus.hashCode() : 0);
            result = 31 * result + (operationalStatus != null ? operationalStatus.hashCode() : 0);
            result = 31 * result + (jFedExperimentState != null ? jFedExperimentState.hashCode() : 0);
            return result;
        }
    }
}
